/**
* Script: query2.1.js
* Written by: Andrew Helenius
* Updated: 1/7/2014
**/

/* supports:
chainable:
 - take(n)     - take the first n results.
 - first(|fn)  - take first satisfying item, or if no function, the very first item.
 - map(fn)     - perform a map operation with fn.
 - filter(fn)  - perform a filter, using fn as the predicate.
 - reject(fn)  - perform the opposite of filter.
 - get(num)    - tries to get the indexed item.
 - uniq()      - filters the results to only unique items.
 - zip(array)  - combines the contents of the array with the current elements.
 - slice(a, b) - returns results between [a, b).

non-chainable: (they perform the query first)
 - each(fn)         - runs the results through the given function.
 - invoke(method)   - runs the results, invoking the named method.
 - toArray()        - returns an array.
 - contains(o|p)    - returns true if something satisfies the predicate or matches the object.
 - some(o|p)        - returns true if something satisfies the predicate or matches the object.
 - indexOf(o)       - returns -1 if item not found, or the index.
 - every(fn)        - checks to see if all items satisfy the predicate.
 - reduce(fn, memo) - reduces the results, starting at memo, or if not, the first item.
 - length()         - returns the overall length.
 - count(p)         - returns the overall number of times the predicate was satisfied.
 - min(rank)        - returns the minimum element using a ranking function as a benchmark.
 - max(rank)        - returns the maximum element using a ranking function as a benchmark.
 - last()           - returns the last result.
 - sample(num)      - selects a random element, up to num of them or once.
 - sort(fn)         - sorts the resulting list with given function, or uses JS default.
 - groupBy(fn)      - returns an array of values grouped by the grouping function.
*/

// optimization idea came from compilation:
var Query2 = (function() {
	var Env = {
		stop: false,
		take: false,
	}

	function WherePoint(fn, reject)
	{
		var where = fn;
		this.next;
		this.func = fn;
		if (reject)
			this.exec = function(item) { if (!where(item)) this.next.exec(item); }
		else
			this.exec = function(item) { if (where(item)) this.next.exec(item); }
	}

	function MapPoint(fn)
	{
		var map = fn;
		this.next;
		this.func = fn;
		this.exec = function(item) { this.next.exec(map(item)); }
	}

	function Map2Point(fn1, fn2)
	{
		var map1 = fn1, map2 = fn2;
		this.next;
		this.exec = function(item) { this.next.exec(map2(map1(item))); }
	}

	function WhereMapPoint(fn1, fn2)
	{
		var where = fn1, map = fn2;
		this.next;
		this.exec = function(item) { if (where(item)) this.next.exec(map(item)); }
	}

	function MapWherePoint(fn1, fn2)
	{
		var where = fn2, map = fn1;
		this.next;
		this.exec = function(item) {
			var i = map(item);
			if (where(i)) this.next.exec(i);
		}
	}

	function Where2Point(fn1, fn2)
	{
		var wh1 = fn1, wh2 = fn2;
		this.next;
		this.exec = function(item) {
			if (wh1(item) && wh2(item)) this.next.exec(item);
		}
	}

	function ZipPoint(array) {
		var a = array, i = 0;
		this.exec = function(item) { this.next.exec([item, a[i++]]); }
	}
	
	function GroupByPoint(fn) {
		var a = {}, group = fn;
		this.exec = function(item) {
			var index = group(item);
			if (!a[index]) a[index] = [item];
			else a[index].push(item);
		}
		this.getGroup = function() { var o = a; a = {}; return o; }
	}

	function SlicePoint(index1, index2) {
		var a = index1, b = index2, i = 0;
		this.exec = function(item) { if (i >= b) Env.stop = true; else if (i >= a) this.next.exec(item); i++; }
	}

	function FirstPoint(fn)
	{
		var func = fn;
		this.next;
		if (func) {
			this.exec = function(item) {
				if (func(item)) Env.stop = true;
				this.next.exec(item);
			}
		}
		else {
			this.exec = function(item) {
				Env.stop = true;
				this.next.exec(item);
			}
		}
	}

	function GetPoint(num) {
		var n = num, c = 0;
		this.next;
		this.exec = function(item) {
			if (c == n) {
				Env.stop = true; c = 0;
				this.next.exec(item);
			} else c++;
		}
	}

	function ContainsPoint(o) {
		var obj = o;
		this.pass = false;
		if (typeof o == "function")
			this.exec = function(item) { if (obj(item)) this.pass = Env.stop = true; };
		else
			this.exec = function(item) { if (item == obj) this.pass = Env.stop = true; };
	}
	
	function EveryPoint(fn) {
		var func = fn;
		this.pass = true;
		this.exec = function(item) { if (!func(item)) { this.pass = false; Env.stop = true; } };
	}

	function IndexOfPoint(o) {
		var obj = o, idx = -1, i = 0;
		this.exec = function(item) { idx++; if (item == obj) { i = idx; Env.stop = true; } };
		this.getIdx = function() { var v = idx != i ? -1 : i; i = idx = -1; return v; };
	}

	function EachPoint(fn) {
		this.exec = fn;
	}
	
	function MaxPoint(fn, min) {
		var rank = fn, value = min ? Number.MAX_VALUE : Number.MIN_VALUE;
		if (min)
			this.exec = function(item) { if (rank(item) < value) value = item; }
		else
			this.exec = function(item) { if (rank(item) > value) value = item; }
		this.getValue = function() { var v = value; value = 0; return v; }
	}
	
	function InvokePoint(method, ctxt) {
		var name = method, ctxt = context;
		if (method) {
			if (context)
				this.exec = function(item) { item[name].call(ctxt); }
			else
				this.exec = function(item) { item[name](); }
		}
		else if (context)
			this.exec = function(item) { item.call(ctxt); }
		else
			this.exec = function(item) { item(); }
	}

	function WhereEachPoint(f1, f2) {
		var where = f1, each = f2;
		this.exec = function(item) { if (where(item)) each(item); }
	}

	function TakePoint(size)
	{
		var i = 0, s = size;
		this.next;
		this.exec = function(item) {
			this.next.exec(item);
			if (++i == s) { Env.stop = true; i = 0; }
		}
	}

	function CountPoint(func)
	{
		var fn = func;
		this.counts = { num: 0, total: 0 };
		this.exec = function(item) { this.counts.total++; if (fn(item)) this.counts.num++; }
		this.getCount = function() { var c = this.counts; this.counts = { num: 0, total: 0 }; return c; }
	}

	function UniqPoint() // still experimental
	{
		var t = [];
		this.next;
		this.exec = function(item) { if (!t[item]) { t[item] = true; this.next.exec(item); } };
	}

	function LengthPoint() {
		var n = 0;
		this.exec = function() { n++; };
		this.num = function() { return n; }
	}

	function ReducePoint(fn, m) {
		var func = fn, memo = m, def = m;
		
		if (m !== undefined)
			this.exec = function(item) { memo = fn(memo, item); };
		else
			this.exec = function(item) {
				memo = def = item;
				this.exec = function(item) { memo = fn(memo, item); };
			}
		this.getMemo = function() { var v = memo; memo = def; return v; }
	}

	// a special kind of end point
	function AllPoint() {
		var i = 0, a = [];
		this.exec = function(item) { a[i++] = item; };
		this.getArray = function() { var arr = a; a = []; i = 0; return arr; };
	}
	
	var array, root, end, prev;
	
	function PushPoint(point) {
		prev = end;
		end.next = point;
		end = point;
	}
	
	function Each(fn) {
		this.run(new EachPoint(fn));
		prev.next = null;
		end = prev;
	}
	
	function Run(point) {
		Env.stop = false;
		if (point) this.pushPoint(point);
		var a = array, l = a.length, i = 0, r = root.next, e = Env;
		if (e.take)
			while (i < l && !e.stop) r.exec(a[i++]);
		else
			while (i < l) r.exec(a[i++]);
	}
	
	function ToArray() {
		var point = new AllPoint();
		this.run(point);
		prev.next = null;
		end = prev;
		return point.getArray();
	}
	
	function Count(fn) {
		var point = new CountPoint(fn);
		this.run(point);
		return point.getCount();
	}
	
	function Length() {
		var point = new LengthPoint();
		this.run(point);
		return point.num();
	}
	
	function Contains(o) {
		Env.take = true;
		var point = new ContainsPoint(o);
		this.run(point);
		return point.pass;
	}
	
	function IndexOf(o) {
		Env.take = true;
		var point = new IndexOfPoint(o);
		this.run(point);
		return point.getIdx();
	}
	
	function GroupBy(fn) {
		var point = new GroupByPoint(fn);
		this.run(point);
		return point.getGroup();
	}
	
	function Every(fn) {
		Env.take = true;
		var point = new EveryPoint(fn);
		this.run(point);
		return point.pass;
	}
	
	function Reduce(agg, memo) {
		var point = new ReducePoint(agg, memo);
		this.run(point);
		return point.getMemo();
	}
		
	function Sample(num) {
		if (!num) num = 1;
		Env.take = true;
		this.pushPoint(new SamplePoint(num));
		return this;
	}
	
	function From(a) {
		Env.stop = false;
		Env.take = false;
		array = a;
		root = { next: null };
		end = root;
		return this;
	}
	
	function Where(func) {
		if (end instanceof WherePoint) {
			var e = new Where2Point(end.func, func);
			prev.next = e;
			end = e;
		}
		else if (end instanceof MapPoint) {
			var e = new MapWherePoint(end.func, func);
			prev.next = e;
			end = e;
		}
		else
			this.pushPoint(new WherePoint(func));
		return this;
	}
	
	function Reject(func) {
		if (end instanceof WherePoint) {
			var e = new Where2Point(end.func, func, true);
			prev.next = e;
			end = e;
		}
		else if (end instanceof MapPoint) {
			var e = new MapWherePoint(end.func, func, true);
			prev.next = e;
			end = e;
		}
		else
			this.pushPoint(new WherePoint(func, true));
		return this;
	}
	
	function Map(func) {
		if (end instanceof WherePoint) {
			var e = new WhereMapPoint(end.func, func);
			prev.next = e;
			end = e;
		}
		else if (end instanceof MapPoint) {
			var e = new Map2Point(end.func, func);
			prev.next = e;
			end = e;
		}
		else
			this.pushPoint(new MapPoint(func));
		return this;
	}
	
	function Max(rank) {
		var point = new MaxPoint(rank);
		this.run(point);
		return point.getValue();
	}
	
	function Min(rank) {
		var point = new MaxPoint(rank, true);
		this.run(point);
		return point.getValue();
	}
	
	function Invoke(name, context) {
		this.pushPoint(new InvokePoint(name, context));
		return this;
	}
	
	function First(fn) {
		Env.take = true;
		this.pushPoint(new FirstPoint(fn));
		return this;
	}
	
	function Zip(array) {
		this.pushPoint(new ZipPoint(array));
		return this;
	}
	
	function Slice(a, b) {
		if (a == 0) return this;
		Env.take = true;
		if (!b) b = Number.MAX_VALUE;
		this.pushPoint(new SlicePoint(a, b));
		return this;
	}
	
	function Last() {
		var a = this.toArray();
		return a[a.length - 1];
	}
	
	function Random(times) {
		if (times === undefined) times = 1;
		var a = this.toArray();
		var samples = [];
		while(times--) {
			var i = Math.floor(Math.random() * a.length);
			samples.push(a[i]);
			a.splice(i, 1);
		}
		return samples;
	}
	
	function Take(n) {
		Env.take = true;
		this.pushPoint(new TakePoint(n));
		return this;
	}
	
	function Get(num) {
		Env.take = true;
		if (num < 0) Env.stop = true;
		else if (num == 0)
			this.pushPoint(new FirstPoint());
		else
			this.pushPoint(new GetPoint(num));
		return this;
	}
	
	function Uniq() {
		this.pushPoint(new UniqPoint());
		return this;
	}
	
	function Sort(f) {
		var v = this.toArray();
		if (f) v.sort(f); else v.sort();
		return v;
	}

	return {
		// pushPoint: adds nodes to parse your functional logic
		pushPoint: PushPoint,
			
		// each: runs the functional logic parser for each element with supplied function.
		each: Each,

		// run: runs the lazy functional logic parser, and adding the final node point.
		run: Run,
		
		// toArray: runs the functional logic parser, returning an array of the results.
		toArray: ToArray,

		// count: counts the number of times the result is matched.
		// count object: { num: num_successes , total: total_elements }
		count: Count,
		
		// length: returns the number of results.
		length: Length,
		size: Length,

		// contains: runs the logic parser, returning true if an item is found.
		contains: Contains,
		
		// invoke: runs all functions or functions with the method name.
		invoke: Invoke,
		
		// some: returns true if something in the list satisfies the predicate.
		some: Contains,
		exists: Contains,
		
		// indexOf: runs the logic, returning -1 if not found, or the index number.
		indexOf: IndexOf,
		
		// groupBy: returns an array of items grouped by the grouping function.
		groupBy: GroupBy,
		
		// every: runs the functional logic parser, checking if each result satisfies the predicate.
		every: Every,
		
		// reduce: runs the functional logic parser, reducing the results.
		reduce: Reduce,
		
		// from: sets up the original array to query from, it will not be modified.
		from: From,
		
		// where: filters the query using the predicate function.
		where: Where,
		filter: Where,
		
		// reject: opposite of filter
		reject: Reject,

		// map: modifies each result in the results list through a function.
		map: Map,
		
		// first: gets the first result, or the first one accepted by the predicate.
		first: First,
		
		// zip: combines elements of the target array with the current results.
		zip: Zip,
		
		// slice: takes only the elements between [a, b). 
		slice: Slice,
		
		// sample: take a random or n random samples from the list.
		sample: Sample,
		
		// last: grab the last item (there is no easy way...)
		last: Last,
		
		// random: get a random item from the results list (there is no easy way...)
		random: Random,
		sample: Random,
		
		// take: filters out everything but the first n items:
		take: Take,
		
		// get: filters to a particular, indexed result.
		get: Get,
		
		// uniq: filters the list to only unique results.
		uniq: Uniq,
		unique: Uniq,

		// sort: sorts the array (WARNING: not lazy... yet).
		sort: Sort,
		
		// max: returns the maximum item
		max: Max,
		
		// min: returns the minimum item
		min: Min
	}
})();