/**
* Script: query2.js
* Written by: Andrew Helenius
* Updated: 1/6/2014
**/

/* supports:
chainable:
 - take(n)     - take the first n results.
 - first(|fn)  - take first satisfying item, or if no function, the very first item.
 - map(fn)     - perform a map operation with fn.
 - where(fn)   - perform a filter using fn as the predicate.
 - get(num)    - tries to get the indexed item.
 - uniq()      - filters the results to only unique items.
 - sort(fn)    - sorts the contents with the function or uses the JS default.
 - zip(array)  - combines the contents of the array with the current elements.
 - slice(a, b) - returns results between [a, b).

non-chainable:
 - each(fn)         - perform the query, running the results through a function.
 - toArray()        - perform the query, returning an array.
 - contains(o|p)    - perform the query, returning true if it satisfies the predicate or matches the object.
 - indexOf(o)       - perform the query, returning -1 if item not found, or the index.
 - every(fn)        - perform the query, checking to see if all items satisfy the predicate.
 - reduce(fn, memo) - perform the query, reducing the results, starting at memo, or if not, the first item.
 - length()         - perform the query, returning the overall length.
 - count(p)         - perform the query, returning the overall number of times the predicate was satisfied.
*/

// optimization idea came from compilation:
var Env = {
	stop: false,
	take: false,
}

function WherePoint(fn)
{
	var where = fn;
	this.next;
	this.func = fn;
	this.exec = function(item) { if (where(item)) this.next.exec(item); }
}

function MapPoint(fn)
{
	var map = fn;
	this.next;
	this.func = fn;
	this.exec = function(item) { this.next.exec(map(item)); }
}

function Map2Point(fn1, fn2)
{
	var map1 = fn1, map2 = fn2;
	this.next;
	this.exec = function(item) { this.next.exec(map2(map1(item))); }
}

function WhereMapPoint(fn1, fn2)
{
	var where = fn1, map = fn2;
	this.next;
	this.exec = function(item) { if (where(item)) this.next.exec(map(item)); }
}

function MapWherePoint(fn1, fn2)
{
	var where = fn2, map = fn1;
	this.next;
	this.exec = function(item) {
		var i = map(item);
		if (where(i)) this.next.exec(i);
	}
}

function Where2Point(fn1, fn2)
{
	var wh1 = fn1, wh2 = fn2;
	this.next;
	this.exec = function(item) {
		if (wh1(item) && wh2(item)) this.next.exec(item);
	}
}

function ZipPoint(array) {
	var a = array, i = 0;
	this.exec = function(item) { this.next.exec([item, a[i++]]); }
}

function SlicePoint(index1, index2) {
	var a = index1, b = index2, i = 0;
	this.exec = function(item) { if (i >= b) Env.stop = true; else if (i >= a) this.next.exec(item); i++; }
}

function FirstPoint(fn)
{
	var func = fn;
	this.next;
	if (func) {
		this.exec = function(item) {
			if (func(item)) Env.stop = true;
			this.next.exec(item);
		}
	}
	else {
		this.exec = function(item) {
			Env.stop = true;
			this.next.exec(item);
		}
	}
}

function GetPoint(num) {
	var n = num, c = -1;
	this.next;
	this.exec = function(item) {
		if (++c == n) {
			Env.stop = true; c = -1;
			this.next.exec(item);
		}
	}
}

function ContainsPoint(o) {
	var obj = o;
	this.pass = false;
	if (typeof o == "function")
		this.exec = function(item) { if (obj(item)) this.pass = Env.stop = true; };
	else
		this.exec = function(item) { if (item == obj) { this.pass = Env.stop = true; } };
}

function EveryPoint(fn) {
	var func = fn;
	this.pass = true;
	this.exec = function(item) { if (!func(item)) { this.pass = false; Env.stop = true; } };
}

function IndexOfPoint(o) {
	var obj = o, idx = -1, i = 0;
	this.exec = function(item) { idx++; if (item == obj) { i = idx; Env.stop = true; } };
	this.getIdx = function() { var v = idx != i ? -1 : i; i = idx = -1; return v; };
}

function EachPoint(fn) {
	this.exec = fn;
}

function WhereEachPoint(f1, f2) {
	var where = f1, each = f2;
	this.exec = function(item) { if (where(item)) each(item); }
}

function TakePoint(size)
{
	var i = 0, s = size;
	this.next;
	this.exec = function(item) {
		this.next.exec(item);
		if (++i == s) { Env.stop = true; i = 0; }
	}
}

function CountPoint(func)
{
	var fn = func;
	this.counts = { num: 0, total: 0 };
	this.exec = function(item) { this.counts.total++; if (fn(item)) this.counts.num++; }
	this.getCount = function() { var c = this.counts; this.counts = { num: 0, total: 0 }; return c; }
}

function UniqPoint() // still experimental
{
	var t = [];
	this.next;
	this.exec = function(item) { if (!t[item]) { t[item] = true; this.next.exec(item); } };
}

function LengthPoint() {
	var n = 0;
	this.exec = function() { n++; };
	this.num = function() { return n; }
}

function ReducePoint(fn, m) {
	var func = fn, memo = m, def = m;
	
	if (m !== undefined)
		this.exec = function(item) { memo = fn(memo, item); };
	else
		this.exec = function(item) {
			memo = def = item;
			this.exec = function(item) { memo = fn(memo, item); };
		}
	this.getMemo = function() { var v = memo; memo = def; return v; }
}

// a special kind of end point
function AllPoint() {
	var i = 0, a = [];
	this.exec = function(item) { a[i++] = item; };
	this.getArray = function() { var arr = a; a = []; i = 0; return arr; };
}

/* VERY EXPERIMENTAL */
var Query2 = {
	array: [],
	root: null,
	end: null,
	prev: null,
	
	/*
		Adds nodes to parse your functional logic
	*/
	pushPoint: function(point) {
		this.prev = this.end;
		this.end.next = point;
		this.end = point;
	},
		
	/*
		each: runs the functional logic parser for each element with supplied function.
	*/
	each: function(fn) {
		this.run(new EachPoint(fn));
		this.prev.next = null;
		this.end = this.prev;
	},

	/*
		run: runs the lazy functional logic parser, and adding the final node point.
	*/
	run: function(point) {
		Env.stop = false;
		if (point) this.pushPoint(point);
		var a = this.array, l = a.length, i = 0, r = this.root.next, e = Env;
		if (e.take)
			while (i < l && !e.stop) r.exec(a[i++]);
		else
			while (i < l) r.exec(a[i++]);
	},
	
	/*
		toArray: runs the functional logic parser, returning an array of the results.
	*/
	toArray: function() {
		var point = new AllPoint();
		this.run(point);
		this.prev.next = null;
		this.end = this.prev;
		return point.getArray();
	},

	/*
		count: counts the number of times the result is matched.
		count object: { num: num_successes , total: total_elements }
	*/
	count: function(fn) {
		var point = new CountPoint(fn);
		this.run(point);
		return point.getCount();
	},
	
	/*
		length: returns the number of results.
	*/
	length: function() {
		var point = new LengthPoint();
		this.run(point);
		return point.num();
	},

	/*
		contains: runs the logic parser, returning true if an item is found.
	*/
	contains: function(o) {
		Env.take = true;
		var point = new ContainsPoint(o);
		this.run(point);
		return point.pass;
	},
	
	/*
		indexOf: runs the logic, returning -1 if not found, or the index number.
	*/
	indexOf: function(o) {
		Env.take = true;
		var point = new IndexOfPoint(o);
		this.run(point);
		return point.getIdx();
	},
	
	/*
		every: runs the functional logic parser, checking if each result satisfies the predicate.
	*/
	every: function(fn) {
		Env.take = true;
		var point = new EveryPoint(fn);
		this.run(point);
		return point.pass;
	},
	
	/*
		reduce: runs the functional logic parser, reducing the results.
	*/
	reduce: function(agg, memo) {
		var point = new ReducePoint(agg, memo);
		this.run(point);
		return point.getMemo();
	},
	
	/*
		from: sets up the original array to query from, it will not be modified.
	*/
	from: function(array) {
		Env.stop = false;
		Env.take = false;
		this.array = array;
		this.root = { next: null };
		this.end = this.root;
		return this;
	},
	
	/*
		where: filters the query using the predicate function.
	*/
	where: function(func) {
		if (this.end instanceof WherePoint) {
			var e = new Where2Point(this.end.func, func);
			this.prev.next = e;
			this.end = e;
		}
		else if (this.end instanceof MapPoint) {
			var e = new MapWherePoint(this.end.func, func);
			this.prev.next = e;
			this.end = e;
		}
		else
			this.pushPoint(new WherePoint(func));
		return this;
	},

	/*
		map: modifies each result in the results list through a function.
	*/
	map: function(func) {
		if (this.end instanceof WherePoint) {
			var e = new WhereMapPoint(this.end.func, func);
			this.prev.next = e;
			this.end = e;
		}
		else if (this.end instanceof MapPoint) {
			var e = new Map2Point(this.end.func, func);
			this.prev.next = e;
			this.end = e;
		}
		else
			this.pushPoint(new MapPoint(func));
		return this;
	},
	
	/*
		first: gets the first result, or the first one accepted by the predicate.
	*/
	first: function(fn) {
		Env.take = true;
		this.pushPoint(new FirstPoint(fn));
		return this;
	},
	
	/*
		zip: combines elements of the target array with the current results.
	*/
	zip: function(array) {
		this.pushPoint(new ZipPoint(array));
		return this;
	},
	
	/*
		slice: takes only the elements between [a, b). 
	*/
	slice: function(a, b) {
		Env.take = true;
		this.pushPoint(new SlicePoint(a, b));
		return this;
	},
	
	/*
		last: grab the last item (there is no easy way...)
	*/
	last: function(array) {
		var a = array;
		if (array === undefined) a = this.toArray();
		return a[a.length - 1];
	},
	
	/*
		random: get a random item from the results list (there is no easy way...)
	*/
	random: function(array) {
		var a = array;
		if (array === undefined) a = this.toArray();
		return a[Math.floor(Math.random() * a.length)];
	},
	
	/*
		take: filters out everything but the first n items:
	*/
	take: function(n) {
		Env.take = true;
		this.pushPoint(new TakePoint(n));
		return this;
	},
	
	/*
		get: filters to a particular, indexed result.
	*/
	get: function(num) {
		Env.take = true;
		if (num < 0) Env.stop = true;
		else if (num == 0)
			this.pushPoint(new FirstPoint());
		else
			this.pushPoint(new GetPoint(num));
		return this;
	},
	
	/*
		uniq: filters the list to only unique results.
	*/
	uniq: function() {
		this.pushPoint(new UniqPoint());
		return this;
	},

	/*
		sort: sorts the array (WARNING: not lazy... yet).
	*/
	sort: function(f) {
		var v = this.toArray();
		if (f) v.sort(f); else v.sort();
		return this.from(v);
	},
	
	/*
		Adds a stored procedure: just like using 'where' but with a predefined predicate and arguments.
	*/
	addProcedure: function(name, func) {
		if (name in this) Abort("QueryError: \"" + name + "\" already a procedure.");
		this[name] = function() {
			var array = [], args = [];
			
			for (var i = 0, l = arguments.length; i < l; i++) {
				args[i + 1] = arguments[i];
			}
			
			for (var i = 0, l = this.results.length; i < l; ++i) {
				args[0] = this.results[i];
				if (func.apply(null, args)) array.push(this.results[i]);
			}
			
			this.results = array;
			return this;
		}.bind(this);
	}
}